from __future__ import annotations

import logging
from typing import Optional, Callable, Any

from zonis import DuplicateRoute

log = logging.getLogger(__name__)
deferred_routes = {}


def route(route_name: Optional[str] = None):
    """Turn an async function into a valid IPC route.

    Parameters
    ----------
    route_name: Optional[str]
        An optional name for this IPC route,
        defaults to the name of the function.

    Raises
    ------
    DuplicateRoute
        A route with this name already exists
    """

    def decorator(func: Callable):
        name = route_name or func.__name__
        if name in deferred_routes:
            raise DuplicateRoute

        deferred_routes[name] = func
        return func

    return decorator


class RouteHandler:
    def __init__(
        self,
    ) -> None:
        self._routes: dict[str, Callable] = {}
        self.__is_open: bool = True
        self.__current_ws = None
        self._instance_mapping: dict[str, Any] = {}

    def route(self, route_name: Optional[str] = None):
        """Turn an async function into a valid IPC route.

        Parameters
        ----------
        route_name: Optional[str]
            An optional name for this IPC route,
            defaults to the name of the function.

        Raises
        ------
        DuplicateRoute
            A route with this name already exists

        Notes
        -----
        If this is a method on a class, you will also
        need to use the ``register_class_instance_for_routes``
        method for this to work as an IPC route.
        """

        def decorator(func: Callable):
            name = route_name or func.__name__
            if name in self._routes:
                raise DuplicateRoute

            log.debug("Registered route %s for class %s", name, self.__class__.__name__)
            self._routes[name] = func
            return func

        return decorator

    def load_routes(self) -> None:
        """Loads all decorated routes."""
        global deferred_routes
        for k, v in deferred_routes.items():
            if k in self._routes:
                raise DuplicateRoute

            log.debug("Registered route %s for class %s", k, self.__class__.__name__)
            self._routes[k] = v
        deferred_routes = {}

    def register_class_instance_for_routes(self, instance, *routes) -> None:
        """Register a class instance for the given route.

        When you turn a method on a class into an IPC route,
        you need to call this method with the instance of the class
        to use as well as the names of the routes for registration to work correctly.

        Parameters
        ----------
        instance
            The class instance the methods live on
        routes
            A list of strings representing the names
            of the IPC routes for this class.
        """
        for r in routes:
            self._instance_mapping[r] = instance



import asyncio
import logging

import signal
from typing import Optional, Dict, Any


from zonis import (
    Packet,
    Router,
    RouteHandler,
    UnknownPacket,
    UnhandledWebsocketType,
)
from zonis.packet import (
    RequestPacket,
    IdentifyDataPacket,
    ClientToServerPacket,
)

log = logging.getLogger(__name__)


class Client(RouteHandler):
    """
    Parameters
    ----------
    reconnect_attempt_count: :class:`int`
        The number of times that the :class:`Client` should
        attempt to reconnect.
    url: :class:`str`
        Defaults to ``ws://localhost``.
    port: Optional[:class:`int`]
        The port that the :class:`Client` should use.
    """

    def __init__(
        self,
        *,
        reconnect_attempt_count: int = 1,
        url: str = "ws://localhost",
        port: Optional[int] = None,
        identifier: str = "DEFAULT",
        secret_key: str = "",
        override_key: Optional[str] = None,
    ) -> None:
        super().__init__()
        url = f"{url}:{port}" if port else url
        url = (
            f"ws://{url}"
            if not url.startswith("ws://") and not url.startswith("wss://")
            else url
        )
        self._url: str = url
        self.identifier: str = identifier
        self._reconnect_attempt_count: int = reconnect_attempt_count
        self._connection_future: asyncio.Future = asyncio.Future()

        self._secret_key: str = secret_key
        self._override_key: str = override_key
        self.__is_open: bool = True
        self.__current_ws = None
        self.__task: Optional[asyncio.Task] = None
        self._instance_mapping: Dict[str, Any] = {}

        self.router: Router = Router(self.identifier, None).register_receiver(
            self._request_handler
        )

        # https://github.com/gearbot/GearBot/blob/live/GearBot/GearBot.py
        try:
            for signame in ("SIGINT", "SIGTERM", "SIGKILL"):
                asyncio.get_event_loop().add_signal_handler(
                    getattr(signal, signame),
                    lambda: asyncio.ensure_future(self.close()),
                )
        except Exception as e:
            pass  # doesn't work on windows

    async def block_until_closed(self):
        """A blocking call which releases when the WS closes."""
        await self.router.block_until_closed()

    async def start(self) -> None:
        """Start the IPC client."""
        self.load_routes()
        await self.router.connect_client(
            self._url,
            idp=IdentifyDataPacket(
                secret_key=self._secret_key, override_key=self._override_key
            ),
        )
        log.info(
            "Successfully connected to the server with identifier %s",
            self.identifier,
        )

    async def close(self) -> None:
        """Stop the IPC client."""
        await self.router.close()
        log.info("Successfully closed the client")

    async def _request_handler(self, packet_data, resolution_handler):
        data: RequestPacket = packet_data["data"]
        route_name = data["route"]
        if route_name not in self._routes:
            await resolution_handler(
                data=Packet(
                    identifier=self.identifier,
                    type="FAILURE_RESPONSE",
                    data=f"{route_name} is not a valid route name.",
                )
            )
            return

        if route_name in self._instance_mapping:
            result = await self._routes[route_name](
                self._instance_mapping[route_name],
                **data["arguments"],
            )
        else:
            result = await self._routes[route_name](**data["arguments"])

        await resolution_handler(
            data=Packet(
                identifier=self.identifier,
                type="SUCCESS_RESPONSE",
                data=result,
            )
        )

    async def request(self, route: str, **kwargs):
        """Make a request to the server"""
        request_future: asyncio.Future = await self.router.send(
            ClientToServerPacket(
                identifier=self.identifier,
                type="CLIENT_REQUEST",
                data=RequestPacket(route=route, arguments=kwargs),
            )
        )
        data: Packet = await request_future
        if "type" not in data:
            log.debug("Failed to resolve packet type for %s", data)
            raise UnknownPacket

        if "data" not in data:
            log.debug(
                "Failed to resolve packet as it was missing the 'data' field: %s",
                data,
            )
            raise UnknownPacket

        if data["type"] != "SUCCESS_RESPONSE":
            raise UnhandledWebsocketType(
                f"Client.request expected a packet of type "
                f"SUCCESS_RESPONSE. Received {data['type']}"
            )

        return data["data"]



from typing import TypedDict, Any, Optional, Literal, Type, Dict

from zonis.exceptions import (
    BaseZonisException,
    DuplicateConnection,
    UnhandledWebsocketType,
)

# Closure codes can be between 3000-4999
custom_close_codes: Dict[int, Type[BaseZonisException]] = {
    4102: DuplicateConnection,
    3001: UnhandledWebsocketType,
}


class Packet(TypedDict):
    data: Any
    type: Literal[
        "IDENTIFY",
        "REQUEST",
        "SUCCESS_RESPONSE",
        "FAILURE_RESPONSE",
        "CLIENT_REQUEST",
        "CLIENT_REQUEST_RESPONSE",
    ]
    identifier: str


class RequestPacket(TypedDict):
    route: str
    arguments: Dict[str, Any]


class IdentifyDataPacket(TypedDict):
    override_key: Optional[str]
    secret_key: str


class IdentifyPacket(TypedDict):
    identifier: str
    type: Literal["IDENTIFY"]
    data: IdentifyDataPacket


class ClientToServerPacket(TypedDict):
    identifier: str
    type: Literal["CLIENT_REQUEST"]
    data: RequestPacket



import asyncio
import logging

log = logging.getLogger(__name__)
background_tasks: set[tuple[str, asyncio.Task]] = set()


def create_task(*args, router_id: str, **kwargs) -> asyncio.Task:
    """Creates a task while maintain a background reference so they dont get de-reffed"""
    task = asyncio.create_task(*args, **kwargs)
    background_tasks.add((router_id, task))
    task.add_done_callback(lambda _: background_tasks.discard((router_id, task)))
    return task



class BaseZonisException(Exception):
    """A base exception handler for Zonis."""

    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = self.__doc__

    def __str__(self):
        return self.message


class MissingReceiveHandler(BaseZonisException):
    """Cannot handle incoming requests due to a lack of receive handlers."""


class DuplicateConnection(BaseZonisException):
    """You have attempted to connect with a duplicated identifier.
    Please try again with a unique one or provide the correct override key."""


class DuplicateRoute(BaseZonisException):
    """You are attempting to register multiple routes with the same name.
    Consider setting the route_name argument to something unique."""


class UnhandledWebsocketType(BaseZonisException):
    """Found a websocket type we can't handle."""


class UnknownRoute(BaseZonisException):
    """The route you requested does not exist."""


class UnknownClient(BaseZonisException):
    """The client you requested is not currently connected."""


class UnknownPacket(BaseZonisException):
    """Router received a packet without a known handler."""


class RequestFailed(BaseZonisException):
    """This request resulted in an error on the end client."""

    def __init__(self, data):
        super().__init__()
        self.response_data = data

    def __str__(self):
        return self.message + "\n\n" + self.response_data



from .ws_impls import WebsocketProtocol, FastAPIWebsockets, Websockets
from .packet import (
    Packet,
    RequestPacket,
    ClientToServerPacket,
    IdentifyPacket,
    IdentifyDataPacket,
)
from .exceptions import *
from .route_registration import RouteHandler, route
from .router import Router
from .server import Server
from .client import Client

__all__ = (
    "WebsocketProtocol",
    "FastAPIWebsockets",
    "Websockets",
    "RouteHandler",
    "Packet",
    "RequestPacket",
    "Server",
    "Client",
    "route",
    "BaseZonisException",
    "DuplicateConnection",
    "DuplicateRoute",
    "UnhandledWebsocketType",
    "ClientToServerPacket",
    "UnknownRoute",
    "UnknownClient",
    "RequestFailed",
    "UnknownPacket",
    "MissingReceiveHandler",
    "Router",
    "IdentifyPacket",
    "IdentifyDataPacket",
)

__version__ = "2.0.0"



from __future__ import annotations

import asyncio
import json
import logging
import secrets
import traceback
import typing as t
from functools import partial

import websockets.client
from websockets.legacy.client import WebSocketClientProtocol

from zonis import (
    UnknownPacket,
    MissingReceiveHandler,
    util,
    WebsocketProtocol,
    Websockets,
)
from zonis.packet import IdentifyPacket, IdentifyDataPacket

log = logging.getLogger(__name__)


class PacketT(t.TypedDict):
    packet_id: str
    type: t.Literal["request", "response"]
    data: dict


class Router:
    """A router class for enabling two-way communication down a single pipe.

    This is built on the architecture ideals that if you want to send
    something down the pipe, then this class can return you the response
    just fine. However, if you wish to receive content for this class
    then you need to register an async method to handle that incoming data
    as this class doesn't have the context required to handle it.
    """

    _EXIT_LITERAL: t.Final = "CLOSE_FUTURE"

    def __init__(
        self,
        identifier: str,
        websocket_connection: WebsocketProtocol | None,
        *,
        retry_on_exception: bool = True,
        max_retries: int = 3,
        default_client_websocket_factory: t.Type[WebsocketProtocol] = Websockets
    ):
        """

        Parameters
        ----------
        identifier: str
            How this connection identifies itself.
        websocket_connection: WebsocketProtocol
            An implementation of the websocket protocol.

            At this stage this only affects the Server,
            Client's will always use Websockets
        retry_on_exception: bool
            On failure, give it another go
        max_retries: int
            Max times to retry on exception

        Warnings
        --------
        websocket_connection should **only ever** be
        None when called from Client. If you ever use this,
        ensure it is passed or set before first usage.
        """
        self._websocket_connection: WebsocketProtocol | None = websocket_connection
        self._default_client_websocket_factory: t.Type[
            WebsocketProtocol
        ] = default_client_websocket_factory
        self._retry_on_exception: bool = retry_on_exception
        self._max_retries: int = max_retries
        self._attempted_connection_count: int = 0
        self.__url: str | None = None
        self.__is_open: bool = True
        self._ws_task: asyncio.Task | None = None
        self._receive_handler = None
        self.identifier: str = identifier
        self._router_id: str = secrets.token_bytes(32).hex()
        self._futures: dict[str, asyncio.Future] = {}
        self._item_queue: asyncio.Queue[
            tuple[t.Literal["SEND_THEN_RECEIVE_RESPONSE"], dict]
            | tuple[t.Literal["SEND_RESPONSE"], dict]
            | tuple[t.Literal["CLOSE_FUTURE"]],
        ] = asyncio.Queue()
        self._connection_future: asyncio.Future | None = None
        self._block_future: asyncio.Future = asyncio.Future()
        self._receive_from_queue_task: asyncio.Task | None = None
        self._receive_from_ws_task: asyncio.Task | None = None

    async def block_until_closed(self):
        """A blocking call which releases when the WS closes.

        Notes
        -----
        This will block even if the
        underlying WS has yet to connect.
        """
        await self._block_future

    async def connect_client(self, url: str, idp: IdentifyDataPacket) -> None:
        """A blocking call which connects the underlying WS."""
        log.debug("Initiating connection to %s with identity %s", url, idp)
        self._connection_future = asyncio.Future()
        self.__url = url
        self._ws_task = util.create_task(
            self._handle_ws(url, idp), router_id=self._router_id
        )

        await self._connection_future

    async def connect_server(self) -> None:
        self._ws_task = util.create_task(self._handle_pipe(), router_id=self._router_id)

    async def close(self) -> None:
        """Close the underlying WS"""
        log.debug("Closing WS connection to %s", self.__url)
        self._item_queue.put_nowait((self._EXIT_LITERAL,))  # type: ignore
        self._check_for_congestion()

        # Let the event loop close the task by giving
        # back control for a no-op. Not sure if this
        # actually then kills the task or not but whatever
        await asyncio.sleep(0)

    async def send(self, data: dict) -> asyncio.Future:
        """Send data down the pipe and receive a response.

        Parameters
        ----------
        data: dict
            The packet to send down the pipe

        Returns
        -------
        asyncio.Future
            This future will contain the returned data
            once it becomes available
        """
        packet_id = secrets.token_bytes(16).hex()
        future: asyncio.Future = asyncio.get_running_loop().create_future()
        self._futures[packet_id] = future
        self._item_queue.put_nowait(
            (
                "SEND_THEN_RECEIVE_RESPONSE",
                {"packet_id": packet_id, "type": "request", "data": data},
            )
        )
        self._check_for_congestion()
        log.debug("Queued item to send as a request for packet %s", packet_id)
        return future

    async def send_response(self, *, packet_id: str, data: dict | None = None):
        """Sends a response to a received request without waiting anything.

        Parameters
        ----------
        packet_id: str
            The packet ID to respond to
        data: dict
            The data being responded with
        """
        if data is None:
            data = {}

        packet: PacketT = {"type": "response", "packet_id": packet_id, "data": data}
        self._item_queue.put_nowait(("SEND_RESPONSE", packet))
        self._check_for_congestion()
        log.debug("Queued item to send as a response for packet %s", packet_id)

    def register_receiver(self, callback) -> Router:
        """Register the provided callback to handle packets
        with the request type.

        Parameters
        ----------
        callback
            The callback to call with the packet data.
        """
        if self._receive_handler is not None:
            log.debug("Router overriding existing receive handler")

        self._receive_handler = callback
        return self

    async def _handle_client_identify(self, idp):
        future = await self.send(
            IdentifyPacket(identifier=self.identifier, type="IDENTIFY", data=idp)
        )
        await future
        self._connection_future.set_result(None)

    async def _handle_ws(self, url, idp):
        """Handles the underlying websocket and distributed callouts.

        This method is called from clients.
        """
        async for websocket in websockets.client.connect(
            url,
            ping_timeout=100000,
        ):
            try:
                websocket = t.cast(WebSocketClientProtocol, websocket)
                self._websocket_connection = self._default_client_websocket_factory(
                    websocket
                )
                # This is a task because sending is done within _handle_pipe
                util.create_task(
                    self._handle_client_identify(idp), router_id=self._router_id
                )
                log.debug("Websocket opened to %s", url)

                result = await self._handle_pipe()
                if result is not None and result == self._EXIT_LITERAL:
                    break

            except Exception as error:
                log.critical("%s", "".join(traceback.format_exception(error)))
            finally:
                self._attempted_connection_count += 1
                if (
                    self._retry_on_exception
                    and self._max_retries < self._attempted_connection_count
                ):
                    log.critical("Exceeded maximum websocket reconnections, exiting...")
                    break

        self._block_future.set_result(None)

    def _ensure_tasks_exist(self):
        if self._receive_from_ws_task is None:
            self._receive_from_ws_task = util.create_task(
                self._websocket_connection.receive(), router_id=self._router_id
            )

        if self._receive_from_queue_task is None:
            self._receive_from_queue_task = util.create_task(
                self._item_queue.get(), router_id=self._router_id
            )

    async def _handle_pipe(self):
        try:
            while self.__is_open:
                self._ensure_tasks_exist()

                done, pending = await asyncio.wait(  # noqa
                    [self._receive_from_ws_task, self._receive_from_queue_task],
                    return_when=asyncio.FIRST_COMPLETED,
                )

                result = done.pop().result()
                if not isinstance(result, tuple):
                    log.debug("< Processing incoming websocket message")
                    # This has been received from the WS
                    self._receive_from_ws_task = None
                    await self._resolve_future(result)
                    continue

                self._receive_from_queue_task = None
                if result[0] == self._EXIT_LITERAL:
                    # Kill the ws connection
                    self._receive_from_ws_task = None

                    self.__is_open = False
                    log.debug("Closing WS connection to %s", self.__url)
                    return self._EXIT_LITERAL

                elif result[0] == "SEND_RESPONSE":
                    log.debug("> Sending response to existing request")
                    await self._dispatch_future(result[1])

                elif result[0] == "SEND_THEN_RECEIVE_RESPONSE":
                    log.debug("> Sending request pending future response")
                    await self._dispatch_future(result[1])

                else:
                    log.critical("Unhandled result: %s", result)
        except Exception as error:
            log.critical("%s", "".join(traceback.format_exception(error)))

    async def _dispatch_future(self, data: dict):
        """Given a future from the queue, send it down the pipe"""
        if "packet_id" not in data:
            log.debug("Failed to resolve send packet as it was missing an id: %s", data)
            raise UnknownPacket

        packet_id = data["packet_id"]
        log.debug("> Sending packet %s", packet_id)
        await self._websocket_connection.send(json.dumps(data))

    async def _resolve_future(self, raw_packet: str | bytes | dict):
        """Given a packet received on the pipe, resolve it's associated future"""
        if isinstance(raw_packet, dict):
            packet: PacketT = raw_packet
        else:
            packet: PacketT = json.loads(raw_packet)

        if "type" not in packet:
            log.debug("Failed to resolve packet type for %s", packet)
            raise UnknownPacket

        if "data" not in packet:
            log.debug(
                "Failed to resolve packet as it was missing the 'data' field: %s",
                packet,
            )
            raise UnknownPacket

        if "packet_id" not in packet:
            log.debug("Failed to resolve packet as it was missing an id: %s", packet)
            raise UnknownPacket

        packet_id = packet["packet_id"]
        packet_data = packet["data"]

        if packet["type"] == "response":
            matching_future: asyncio.Future | None = self._futures.pop(packet_id, None)
            if matching_future is None:
                raise UnknownPacket

            matching_future.set_result(packet_data)

        else:
            # This is a request to propagate up to
            # the incoming request handler method
            if self._receive_handler is None:
                raise MissingReceiveHandler

            # Give the callee the incoming data
            # and function to call to return a response
            util.create_task(
                self._receive_handler(
                    packet_data,
                    partial(self.send_response, packet_id=packet["packet_id"]),
                ),
                router_id=self._router_id,
            )

    def _check_for_congestion(self):
        """Checks if the queue appears congested and alerts"""
        # 50 is an arb number I picked out of thin air
        # Would love feedback on this size if it affects you a lot
        if self._item_queue.qsize() > 50:
            log.warning("Router queue size is over 50, possible congestion detected")



import logging
import secrets
import traceback
from typing import Dict, Literal, Any, cast, Optional, Callable

from websockets.exceptions import ConnectionClosedOK, ConnectionClosedError

from zonis import (
    Packet,
    UnknownClient,
    RequestFailed,
    BaseZonisException,
    DuplicateConnection,
    Router,
    RouteHandler,
    FastAPIWebsockets,
)
from zonis.packet import RequestPacket, IdentifyPacket
from zonis.router import PacketT

log = logging.getLogger(__name__)


class Server(RouteHandler):
    """
    Parameters
    ----------
    using_fastapi_websockets: :class:`bool`
        Defaults to ``False``.
    override_key: Optional[:class:`str`]
    secret_key: :class:`str`
        Defaults to an emptry string.
    """

    def __init__(
        self,
        *,
        using_fastapi_websockets: bool = False,
        override_key: Optional[str] = None,
        secret_key: str = "",
    ) -> None:
        super().__init__()
        self._connections: dict[str, Router] = {}
        self._secret_key: str = secret_key
        self._override_key: Optional[str] = (
            override_key if override_key is not None else secrets.token_hex(64)
        )
        self.using_fastapi_websockets: bool = using_fastapi_websockets

        self.__is_open = True

    async def disconnect(self, identifier: str) -> None:
        """Disconnect a client connection.

        Parameters
        ----------
        identifier: str
            The client identifier to disconnect

        Notes
        -----
        This doesn't yet tell the client to stop
        gracefully, this just removes it from our store.

        """
        router = self._connections.pop(identifier)
        await router.close()

    async def request(
        self, route: str, *, client_identifier: str = "DEFAULT", **kwargs
    ) -> Any:
        """Make a request to the provided IPC client.

        Parameters
        ----------
        route: str
            The IPC route to call.
        client_identifier: Optional[str]
            The client to make a request to.

            This only applies in many to one setups
            or if you changed the default identifier.
        kwargs
            All the arguments you wish to invoke the IPC route with.

        Returns
        -------
        Any
            The data the IPC route returned.

        Raises
        ------
        RequestFailed
            The IPC request failed.
        """
        conn = self._connections.get(client_identifier)
        if not conn:
            raise UnknownClient

        request_future = await conn.send(
            Packet(
                identifier=client_identifier,
                type="REQUEST",
                data=RequestPacket(route=route, arguments=kwargs),
            )
        )
        packet = await request_future
        if packet["type"] == "FAILURE_RESPONSE":
            raise RequestFailed(packet["data"])

        return packet["data"]

    async def request_all(self, route: str, **kwargs) -> Dict[str, Any]:
        """Issue a request to connected IPC clients.

        Parameters
        ----------
        route: str
            The IPC route to call.
        kwargs
            All the arguments you wish to invoke the IPC route with.

        Returns
        -------
        Dict[str, Any]
            A dictionary where the keys are the client
            identifiers and the values are the returned data.

            The data could also be an instance of :py:class:RequestFailed:
        """
        results: Dict[str, Any] = {}

        for i, conn in self._connections.items():
            try:
                request_future = await conn.send(
                    Packet(
                        identifier=i,
                        type="REQUEST",
                        data=RequestPacket(route=route, arguments=kwargs),
                    )
                )
                packet: Packet = await request_future
                if packet["type"] == "FAILURE_RESPONSE":
                    results[i] = RequestFailed(packet["data"])
                else:
                    results[i] = packet["data"]
            except ConnectionClosedOK as e:
                results[i] = RequestFailed("Connection Closed")
                log.error(
                    "request_all connection closed: %s, %s",
                    i,
                    "".join(traceback.format_exception(e)),
                )
            except ConnectionClosedError as e:
                results[i] = RequestFailed(
                    f"Connection closed with error: {e.code}|{e.reason}"
                )
                log.error(
                    "request_all connection closed with error: %s, %s",
                    i,
                    "".join(traceback.format_exception(e)),
                )
            except Exception as e:
                results[i] = RequestFailed("Request failed.")
                log.error(
                    "request_all connection threw: %s, %s",
                    i,
                    "".join(traceback.format_exception(e)),
                )

        return results

    async def parse_identify(self, packet: PacketT, websocket) -> str:
        """Parse a packet to establish a new valid client connection.

        Parameters
        ----------
        packet: Packet
            The packet to read
        websocket
            The websocket this connection is using

        Returns
        -------
        str
            The established clients identifier

        Raises
        ------
        BaseZonisException
            Unexpected WS issue
        DuplicateConnection
            Duplicate connection without override keys
        """
        raw_packet = packet
        packet = raw_packet["data"]
        identifier: str = packet.get("identifier")
        try:
            ws_type: Literal["IDENTIFY"] = packet["type"]
            if ws_type != "IDENTIFY":
                await websocket.close(
                    code=4101, reason=f"Expected IDENTIFY, received {ws_type}"
                )
                raise BaseZonisException(
                    f"Unexpected ws response type, expected IDENTIFY, received {ws_type}"
                )

            packet: IdentifyPacket = cast(IdentifyPacket, packet)
            secret_key = packet["data"]["secret_key"]
            if secret_key != self._secret_key:
                await websocket.close(code=4100, reason=f"Invalid secret key.")
                raise BaseZonisException(
                    f"Client attempted to connect with an incorrect secret key."
                )

            override_key = packet["data"].get("override_key")
            if identifier in self._connections and (
                not override_key or override_key != self._override_key
            ):
                await websocket.close(
                    code=4102, reason="Duplicate identifier on IDENTIFY"
                )
                raise DuplicateConnection("Identify failed.")

            router: Router = Router(identifier, FastAPIWebsockets(websocket))
            router.register_receiver(callback=self._request_handler)

            await router.connect_server()
            self._connections[identifier] = router
            await router.send_response(
                packet_id=raw_packet["packet_id"],
                data=Packet(identifier=identifier, type="IDENTIFY", data=None),
            )
            return identifier
        except Exception as e:
            self._connections.pop(identifier, None)
            raise BaseZonisException("Identify failed") from e

    async def _request_handler(self, packet_data, resolution_handler):
        data: RequestPacket = packet_data["data"]
        route_name = data["route"]
        if route_name not in self._routes:
            await resolution_handler(
                data=Packet(
                    identifier="SERVER",
                    type="FAILURE_RESPONSE",
                    data=f"{route_name} is not a valid route name.",
                )
            )
            return

        if route_name in self._instance_mapping:
            result = await self._routes[route_name](
                self._instance_mapping[route_name],
                **data["arguments"],
            )
        else:
            result = await self._routes[route_name](**data["arguments"])

        await resolution_handler(
            data=Packet(
                identifier="SERVER",
                type="SUCCESS_RESPONSE",
                data=result,
            )
        )



from typing import Protocol, runtime_checkable


@runtime_checkable
class WebsocketProtocol(Protocol):
    """The protocol underlying websockets should be exposed via."""

    def __init__(self, ws):
        raise NotImplementedError

    async def send(self, content: str) -> None:
        """Send a message down the wire.

        Parameters
        ----------
        content: str
            The content to send down the wire.
        """
        raise NotImplementedError

    async def receive(self) -> str | bytes | dict:
        """Listen for a message on the wire.

        Returns
        -------
        str|bytes|dict
            The content received
        """
        raise NotImplementedError


class FastAPIWebsockets:
    """A websocket implementation wrapping FastAPI websockets"""

    def __init__(self, ws):
        self.ws = ws

    async def send(self, content: str) -> None:
        await self.ws.send_text(content)

    async def receive(self) -> str | bytes | dict:
        return await self.ws.receive_text()


class Websockets:
    """A websocket implementation wrapping the websockets library"""

    def __init__(self, ws):
        self.ws = ws

    async def send(self, content: str) -> None:
        await self.ws.send(content)

    async def receive(self) -> str | bytes | dict:
        return await self.ws.recv()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zonis']

package_data = \
{'': ['*']}

install_requires = \
['websockets>=10.4,<11.0']

setup_kwargs = {
    'name': 'zonis',
    'version': '2.0.0',
    'description': 'Agnostic IPC for Python programs ',
    'long_description': 'Zonis\n---\n\nA coro based callback system for many to one IPC setups.\n\n`pip install zonis`\n\n\n=======\nSee the [examples](https://github.com/Skelmis/Zonis/tree/master/examples) for simple use cases.\n___\n\n## Build the docs locally\n\nIf you want to build and run the docs locally using sphinx run\n```\nsphinx-autobuild -a docs docs/_build/html --watch zonis\n```\n\nthis will build the docs and start a local server; additionally it will listed for changes to the source directory ``zonis`` and to the docs source directory ``docs/``.\nYou can find the builded files at ``docs/_build``.\n',
    'author': 'skelmis',
    'author_email': 'skelmis.craft@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
